INFRA = "m5.4xlarge"
REPO_OWNER = "OpenMined"
REPO_NAME = "syft-enclave"
REF = "dev"
REGION = "us-west-2"
VCS = "github"
VISIBILITY = "private"
REF_TYPE = "branch"
LOCAL_MODE = True
DOMAIN_CONNECTION_PORT = 3030
